/**
 * Copyright (c) 2014 - 2017, Nordic Semiconductor ASA
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 * 
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 * 
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/** @file
 *
 * @brief Generic Health Sensor Service Sample Application main file.
 *
 */

#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include "nordic_common.h"
#include "nrf.h"
#include "nrf_delay.h"
#include "app_error.h"
#include "nrf_gpio.h"
#include "app_timer.h"
#include "ble.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "ble_conn_params.h"
#include "boards.h"

#include "nrf_sdm.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_strerror.h"

/*
#if defined(APP_USBD_ENABLED) && APP_USBD_ENABLED
    #include "app_usbd_serial_num.h"
    #ifdef BOARD_PCA10059
        #include "nrf_dfu_trigger_usb.h"
    #endif
    #include "app_usbd.h"
#endif
*/

#include "app_timer.h"
#include "app_gpiote.h"
#include "bsp.h"
#include "bsp_btn_ble.h"
//#include "app_uart.h"
#include "ble_conn_state.h"

#include "ble_gap.h"
#include "ble_types.h"
#include "ble_gatts.h"
#include "msmt_queue.h"
#include "GhsControlStructs.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

// lint -e553
#ifdef SVCALL_AS_NORMAL_FUNCTION
#include "ser_phy_debug_app.h"
#endif

#define APP_GPIOTE_MAX_USERS            1                                           /**< Maximum number of users of the GPIOTE handler. */
#define BUTTON_DETECTION_DELAY          APP_TIMER_TICKS(50)                         /**< Delay from a GPIOTE event until a button is reported as pushed (in number of timer ticks). */


#define IS_SRVC_CHANGED_CHARACT_PRESENT 0                                           /**< Include or not the service_changed characteristic. if not enabled, the server's database cannot be changed for the lifetime of the device*/

#define APP_BLE_OBSERVER_PRIO           3                                           /**< Application's BLE observer priority. You shouldn't need to modify this value. */
#define APP_BLE_CONN_CFG_TAG            1                                           /**< A tag identifying the SoftDevice BLE configuration. */

#define APP_ADV_INTERVAL                40                                          /**< The advertising interval (in units of 0.625 ms. This value corresponds to 25 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      180                                         /**< The advertising timeout in units of seconds. */

#define APP_TIMER_PRESCALER             0                                           /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_OP_QUEUE_SIZE         4                                           /**< Size of timer operation queues. */

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(500, UNIT_1_25_MS)            /**< Minimum acceptable connection interval (0.5 seconds) */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(1000, UNIT_1_25_MS)           /**< Maximum acceptable connection interval (1 second). */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory timeout (4 seconds). */

#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000)                       /**< Time from initiating event (connect or start of indication) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(3000)                       /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */

#define GHS_COMMAND_DELAY               APP_TIMER_TICKS(50)
#define GHS_PLETH_DELAY                 APP_TIMER_TICKS(200)
#define CMD_SENSOR_TIME                 APP_TIMER_TICKS(1000)

#define SEC_PARAM_BOND                  1                                           /**< Perform bonding. */
#define SEC_PARAM_MITM                  0                                           /**< Man In The Middle protection not required. */
#define SEC_PARAM_LESC                  0                                           /**< LE Secure Connections not enabled. */
#define SEC_PARAM_KEYPRESS              0                                           /**< Keypress notifications not enabled. */
#define SEC_PARAM_IO_CAPABILITIES       BLE_GAP_IO_CAPS_NONE                        /**< No I/O capabilities. */
#define SEC_PARAM_OOB                   0                                           /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE          7                                           /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE          16                                          /**< Maximum encryption key size. */

#define DEAD_BEEF                       0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

#define APP_FEATURE_NOT_SUPPORTED       BLE_GATT_STATUS_ATTERR_APP_BEGIN + 2        /**< Reply when unsupported features are requested. */

static unsigned long elapsedTimeStart;
static unsigned long elapsedCountStart;

static unsigned short mtu_size = BLE_GATT_ATT_MTU_DEFAULT;

static bool flash_write_needed          = true;

APP_TIMER_DEF(m_ghs_disconnect_timer_id);      /**< disconnect handler */
APP_TIMER_DEF(m_ghs_flash_write_timer_id);
APP_TIMER_DEF(m_app_dummy_timer_id);

static uint8_t                  m_adv_handle                    = 0;  // For advertisments
static uint32_t                 m_config_id                     = 1;  // For advertisments

static uint16_t                 m_connection_handle             = BLE_CONN_HANDLE_INVALID;     /**< Handle of the current connection. */
static uint16_t                 m_dev_info_service_handle       = BLE_GATT_HANDLE_INVALID;
static uint16_t                 m_cyc_pwr_service_handle        = BLE_GATT_HANDLE_INVALID;
static uint16_t                 m_heart_rate_service_handle     = BLE_GATT_HANDLE_INVALID;
static ble_gatts_char_handles_t m_cyc_pwr_msmt_char_handle;
static ble_gatts_char_handles_t m_cyc_pwr_feature_char_handle;
static ble_gatts_char_handles_t m_cyc_pwr_sensor_loc_char_handle;
static ble_gatts_char_handles_t m_heart_rate_msmt_char_handle;
static unsigned char            feature[4]                      = {8, 0, 0, 0};  // supports crank rev data
static unsigned char            sensorLocation[1]               = {FRONT_WHEEL};
static unsigned char            cccdSet[2]                      = {0, 0};
static unsigned short           noOfCccds                       = 2;
static bool                     hasEncrypted                    = false;

static bool                     cycle_abort                     = false;
static ble_gap_sec_params_t     own_sec_params;
static ble_gap_sec_params_t     peer_sec_params;
static ble_gap_sec_keyset_t     keys;
static uint16_t                 saveDataLength                  = 0;
static uint8_t                  *saveDataBuffer                 = NULL;
static volatile bool            restartAdv                      = false;
static unsigned short           mask                            = COUNT_MASK;
ble_gap_conn_params_t           gap_conn_params;


nrf_mutex_t p_mutex;
nrf_mutex_t q_mutex;

__ALIGN(4) uint8_t *evt_buf;
__ALIGN(4) uint8_t *evt_buf2;

s_Queue *queue;

#define BLE_APPEARANCE BLE_APPEARANCE_CYCLING_POWER_SENSOR
static unsigned char bluetoothAddress[]         = {0x02, 0xF1, 0x02, 0xE3, 0x4D, 0xA9}; // A9:4D:E3:02:F1:01 in little endian
static unsigned char systemId[]                 = {0x02, 0xF1, 0x02, 0xFF, 0xFF, 0xE3, 0x4D, 0xA9};

static char *DEVICE_NAME                        = APP_NAME;
static char *MODEL_NUMBER                       = "LNI-M3-001";
static char *MANUFACTURER_NAME                  = "LNI Health, Inc";

static unsigned short cumulative_crank_revs     = 0;
static unsigned long adv_count                  = 1;
static volatile bool first_keiser_adv_found     = false;
static unsigned short time_of_prev_last_crank   = 0;
unsigned short heart_rate                       = 0;
unsigned short power                            = 0;
unsigned short rpm                              = 0;

static unsigned char cycle_power_msmt_char[] = 
{
    CP_MSMT_FLAGS_HAS_CRANK_REVS,       // flags
    0x00,
    0x00,       // instantaneous power in watts resolution 1
    0x00,
    0x00,       // cumulative crank revolutions
    0x00,
    0x00,       // time in units of 1/1024 seconds since last crank detection
    0x00
};

static unsigned char heart_rate_msmt_char[] = 
{
    HEART_RATE_MSMT_FLAGS_TWO_BYTES,       // flags indicate heart rate is two bytes in size
    0x00,
    0x00        // instantaneous heart rate bpm
};
// For the scanner
static ble_gap_scan_params_t scan_params;
static ble_data_t scan_buffer_t;
static unsigned char scan_buffer[31];

int fourByteEncode(unsigned char* msmtBuf, int index, unsigned long value);

/**@brief Function for initializing services that will be used by the application.
 *
 * Create the specified primary standard BTLE service and add it to GATT
 *
 * @param [IN] serviceHandle pointer to a variable to take the created service's handle
 * @param the 16-bit service value, so it needs to be a BTLE standardized service
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
ret_code_t createPrimaryService(unsigned short* serviceHandle, unsigned short serviceUuid);

/**@brief Function for adding the a standard BTLE characteristic [one identified by a standard 16-bit UUID value defined by BT SIG).
 *
 * @param serviceHandle: the handle of the service to which the characteristic belongs
 * @param charHandles: a pointer to a struct that takes the handles of the created characteristic (declaration, value, and descriptor - if present)
 * @param uuid: the 16-bit UUID of the characteristic
 * @param hasCCCD: true if the characteristic is to include a CCCD
 * @param indicate: true if indicate, otherwise notify. Ignored if hasCCCD is false
 * @param valueLength: length of the initial characteristic value. Ignored if 0 or value is NULL
 * @param value: initial value. Ignored if NULL or valueLength is 0
 * @param trapRead if true, a read of the attribute and/or its descriptor will signal the BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST event
 * @param cccdWritePermission: the security level of the CCCD write operation (ignored if no CCCD)
 * @param attrReadPermission: the security level of the characteristic value read operation. If security mode is 0 no read at all.
 * @param attrWritePermission: the security level of the characteristic value write operation. If security mode is 0 no write at all.
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
ret_code_t createStandardCharacteristic(
    uint16_t serviceHandle,
    ble_gatts_char_handles_t* charHandles,
    uint16_t uuid,
    bool hasCCCD,
    bool indicate,
    uint16_t valueLength,
    uint8_t *value,
    bool trapRead,
    ble_gap_conn_sec_mode_t cccdWritePermission,
    ble_gap_conn_sec_mode_t attrReadPermission,
    ble_gap_conn_sec_mode_t attrWritePermission,
    bool isStatic);

/**
 * Method to generate a Device Information Service
 * @param [out]serviceHandle: a pointer to the service handle for the Device Information Service.
 * @param disCharValues: pointer to a struct containing the characteristic values in the service. If NULL that characteristic
 *        will not be created.
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
unsigned long createDeviceInformationService(unsigned short int* serviceHandle, s_DisCharValues* disCharValues);

/**
 * Method saves bonding data to a file
 * @param keys of the peripheral
 * @param saveDataBuffer a buffer to hold the data obtained from the sd_ble_gatts_sys_attr_get() method
 * @param saveDataLength the length of the data obtained from the sd_ble_gatts_sys_attr_get()
 * @param cccdSet a pointer to the list of enabled states of the characteristics
 * @param noOfCccds a pointer to the number of Cccds that can be enabled in the specialization
 */
void saveKeysToFlash(ble_gap_sec_keyset_t* keys, 
    unsigned char** saveDataBuffer, unsigned short int* saveDataLength, unsigned char* cccdSet,
    unsigned short *noOfCccds);

/**
 * Method loads saved bonding data from a file
 * @param keys the peripheral
 * @param saveDataBuffer a buffer to hold the data obtained from the sd_ble_gatts_sys_attr_get() method
 * @param saveDataLength the length of the data obtained from the sd_ble_gatts_sys_attr_get()
 * @param cccdState a pointer to the list of enabled state of the characteristics
 * @param noOfCccds a pointer to the number of Cccds that can be enabled in the specialization
 */
void loadKeysFromFlash(ble_gap_sec_keyset_t* keys, 
    unsigned char** saveDataBuffer, unsigned short int* saveDataLength, unsigned char* cccdSet,
    unsigned short *noOfCccds);


void allocateMemoryForSecurityKeys(ble_gap_sec_keyset_t* keys);
void freeMemoryForSecurityKeys(ble_gap_sec_keyset_t* keys);
char* byteToHex(unsigned char* bytes, char* buffer, char* separator, unsigned long count);

/*
 This GHS sends indications on two characteristics. These variables tell the indicate data method
 which handles (characteristics) to send indications on.
 When the PHD receives a command, the GHS will do one of two things:
    1. indicate a response on the control point characteristic indicating the command has been handled.
       The PHG may send the next command.
    2. indicate a one or more responses on the response characteristic (the response may be fragmented over
       several indications, such as the systemInfo and the measurement groups)
       then indicate a response on the control point characteristic - usually this means the PHD is done
       handling the command, but it may also indicate a measurement group has been completed but more are
       to come. In that case the PHG just handles the measurement group and waits for more. It shall NOT
       send another command. The PHD will send a final response indicating the measurement groups are completed,
       ending the command. Now the PHG may send the next command.
       
       In the live data situation, the command never completes. The PHD will just send measurement groups
       on the response characteristic followed by group finished response on the control point and then
       continue with the next. When the PHD has no more data to send it disconnects. 
 */

static volatile bool send_flag = false;

const char * getErrorString(unsigned long err_code)
{
    char const * p_desc = nrf_strerror_find(err_code);
    return (p_desc != NULL) ? p_desc : "Unknown";

}

// ------------- Current time and time info

static unsigned long getTicksMillis()
{
    unsigned long long count = (unsigned long long)(app_timer_cnt_get() - elapsedTimeStart);
    return (unsigned long)((count * 10000 + 5) / 327680);
}

// Units are 32768 counts per second
// This method get ticks in units of 1024ths of a second
static unsigned long getTicks()
{
    return ((app_timer_cnt_get() >> 5) - elapsedCountStart);
}

static void sec_params_init(void)
{
    own_sec_params.lesc             = SEC_PARAM_LESC;
    own_sec_params.keypress         = SEC_PARAM_KEYPRESS;
    own_sec_params.bond             = SEC_PARAM_BOND;
    own_sec_params.mitm             = SEC_PARAM_MITM;
    own_sec_params.io_caps          = SEC_PARAM_IO_CAPABILITIES;
    own_sec_params.oob              = SEC_PARAM_OOB;
    own_sec_params.min_key_size     = SEC_PARAM_MIN_KEY_SIZE;
    own_sec_params.max_key_size     = SEC_PARAM_MAX_KEY_SIZE;
    own_sec_params.kdist_peer.enc   = 1;     // peer to send their encryption key;
    own_sec_params.kdist_peer.id    = 1;     // peer to send its IRK
    own_sec_params.kdist_peer.link  = 0;     // peer to derive link key from LTK - This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
    own_sec_params.kdist_peer.sign  = 0;     // peer to send signature resolving key - This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
    own_sec_params.kdist_own.enc    = 1;     // We distribute our LTK
    own_sec_params.kdist_own.id     = 1;     // We distribute our IRK (even though we don't use resolvable random addressing
    own_sec_params.kdist_own.link   = 0;     // we are to derive link key from LTK - This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
    own_sec_params.kdist_own.sign   = 0;     // We to send signature resolving key - This must be 0 or you will get an NRF_ERROR_NOT_SUPPORTED error
}

#define ADVERTISING_INTERVAL_40_MS 64  /**< 0.625 ms = 40 ms */
#define BUFFER_SIZE 30

static ble_gap_adv_params_t     m_adv_params;
static ble_gap_adv_data_t       m_adv_data;
static uint8_t                  data_buffer[BUFFER_SIZE];
static uint8_t                  scan_rsp_buffer[BUFFER_SIZE];
ret_code_t advertisement_ghs_set(uint8_t *advHandle)
{
    ret_code_t error_code;
    uint8_t  index = 0;
    uint8_t  scan_index = 0;
    char printBuf[256];
    memset(data_buffer, 0, sizeof(data_buffer));
    memset(scan_rsp_buffer, 0, sizeof(scan_rsp_buffer));

    const uint8_t name_length = (uint8_t)strlen(DEVICE_NAME);

    // Set the flags - important to indicate no EDR/BR support for some Androids
    // Actually by spec one is SUPPOSED to set one of the discoverable flags to be discoverable.
    data_buffer[index++] = 2;        // length of flags data type
    data_buffer[index++] = BLE_GAP_AD_TYPE_FLAGS;
    data_buffer[index++] = BLE_GAP_ADV_FLAG_LE_GENERAL_DISC_MODE | BLE_GAP_ADV_FLAG_BR_EDR_NOT_SUPPORTED;    // No EDR/BR support and general discoverable

    scan_rsp_buffer[scan_index++] = 3;        // length of appearance data type
    scan_rsp_buffer[scan_index++] = BLE_GAP_AD_TYPE_APPEARANCE;
    scan_rsp_buffer[scan_index++] = (BLE_APPEARANCE & 0xFF);
    scan_rsp_buffer[scan_index++] = ((BLE_APPEARANCE >> 8) & 0xFF);

    // Set the device name.
    if (DEVICE_NAME != NULL)
    {
        data_buffer[index++] = name_length + 1; // Device name + data type
        data_buffer[index++] = BLE_GAP_AD_TYPE_COMPLETE_LOCAL_NAME;
        memcpy((char*)&data_buffer[index], DEVICE_NAME, name_length);
        index += name_length;
    }

    // Set the device's service.
    data_buffer[index++] = 1 + 4;
    data_buffer[index++] = BLE_GAP_AD_TYPE_16BIT_SERVICE_UUID_MORE_AVAILABLE;
    // Store in little-endian format.
    data_buffer[index++] = BTLE_CYCLE_POWER_SERVICE & 0xFF;
    data_buffer[index++] = (BTLE_CYCLE_POWER_SERVICE & 0xFF00) >> 8;
    data_buffer[index++] = BTLE_HEART_RATE_SERVICE & 0xFF;
    data_buffer[index++] = (BTLE_HEART_RATE_SERVICE & 0xFF00) >> 8;

    ble_gap_adv_properties_t adv_properties;
    adv_properties.type = BLE_GAP_ADV_TYPE_CONNECTABLE_SCANNABLE_UNDIRECTED;
    adv_properties.anonymous = 0;
    adv_properties.include_tx_power = 0;

    memset(&m_adv_params, 0, sizeof(ble_gap_adv_params_t));
    m_adv_params.properties = adv_properties;
    m_adv_params.filter_policy = BLE_GAP_ADV_FP_ANY;
    m_adv_params.duration = BLE_GAP_ADV_TIMEOUT_GENERAL_UNLIMITED;
    m_adv_params.p_peer_addr = NULL;
    m_adv_params.interval = ADVERTISING_INTERVAL_40_MS;
    m_adv_params.max_adv_evts = 0;
    m_adv_params.primary_phy = BLE_GAP_PHY_AUTO;
    m_adv_params.secondary_phy = BLE_GAP_PHY_AUTO;
    m_adv_params.channel_mask[0] = 0;
    m_adv_params.channel_mask[1] = 0;
    m_adv_params.channel_mask[2] = 0;
    m_adv_params.channel_mask[3] = 0;
    m_adv_params.channel_mask[4] = 0;

    memset(&m_adv_data, 0, sizeof(m_adv_data));

    m_adv_data.adv_data.p_data = data_buffer;
    m_adv_data.adv_data.len = index;
    m_adv_data.scan_rsp_data.p_data = scan_rsp_buffer;
    m_adv_data.scan_rsp_data.len = scan_index;

    memset(printBuf, 0, 256);
    NRF_LOG_INFO("Advertisement is %s", (uint32_t)byteToHex(data_buffer, printBuf, " ", index));
    NRF_LOG_FLUSH();
    if (scan_index > 0)
    {
        memset(printBuf, 0, 256);
        NRF_LOG_INFO("Scan response is: %s", (uint32_t)byteToHex(scan_rsp_buffer, printBuf, " ", scan_index));
        NRF_LOG_FLUSH();
    }

    error_code = sd_ble_gap_adv_set_configure(advHandle, &m_adv_data, &m_adv_params);

    if (error_code != NRF_SUCCESS)
    {
        NRF_LOG_ERROR("Failed to set advertisement data. Error: %s (0x%02X)", getErrorString(error_code), error_code);
        return error_code;
    }
    return NRF_SUCCESS;
}

static uint32_t advertising_start()
{
    ret_code_t             error_code;

    error_code = sd_ble_gap_adv_start(m_adv_handle, m_config_id);

    if (error_code != NRF_SUCCESS)
    {
        NRF_LOG_ERROR("Failed to start advertising. Error: %s (0x%02X)", getErrorString(error_code), error_code);
        return error_code;
    }

    NRF_LOG_INFO("Started advertising at time %u", getTicksMillis());
    return NRF_SUCCESS;
}

/**@brief Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num   Line number of the failing ASSERT call.
 * @param[in] file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
    while(true)
    {
        bsp_board_led_invert(RED_LED);
        nrf_delay_ms(500);
    }
}

static void ble_disconnected_handler(void * p_context)
{
    UNUSED_PARAMETER(p_context);
    ret_code_t  err_code;
    
    if (m_connection_handle != BLE_CONN_HANDLE_INVALID)
    {
        cycle_abort = true;
        err_code = sd_ble_gap_disconnect(m_connection_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
        APP_ERROR_CHECK(err_code);
    }
}


void send_zero_values(void * p_context)
{
    UNUSED_PARAMETER(p_context);
    if (!first_keiser_adv_found && m_connection_handle != BLE_CONN_HANDLE_INVALID)
    {
        s_PowerMsmt powerMsmt;
        powerMsmt.hr = 0;
        powerMsmt.power = 0;
        powerMsmt.rpm  = 0;
        powerMsmt.crankTime = getTicks();
        bsp_board_led_on(RED_LED);
        NRF_LOG_INFO("Zero-valued msmt added to queue");
        enqueue(queue, &powerMsmt, sizeof(s_PowerMsmt));
    }
}

static void write_flash(void * p_context);
static void timers_init(void)
{
    ret_code_t err_code;

    // Initialize timer module, making it use the scheduler.
    err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_ghs_disconnect_timer_id,
                            APP_TIMER_MODE_SINGLE_SHOT,
                            ble_disconnected_handler);

    err_code = app_timer_create(&m_ghs_flash_write_timer_id,
                            APP_TIMER_MODE_SINGLE_SHOT,
                            write_flash);
    err_code = app_timer_create(&m_app_dummy_timer_id,
                            APP_TIMER_MODE_REPEATED,
                            send_zero_values);
    APP_ERROR_CHECK(err_code);

}


/**@brief Function for the GAP initialization.
 *
 * @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 *          device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void)
{
    ret_code_t              err_code;
    ble_gap_conn_params_t   gap_conn_params;
    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (const uint8_t *)DEVICE_NAME,
                                          strlen(DEVICE_NAME));
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE);
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}

/**@brief Function for handling the Connection Parameters Module.
 *
 * @details This function will be called for all events in the Connection Parameters Module which
 *          are passed to the application.
 *          @note All this function does is to disconnect. This could have been done by simply
 *                setting the disconnect_on_fail config parameter, but instead we use the event
 *                handler mechanism to demonstrate its use.
 *
 * @param[in] p_evt  Event received from the Connection Parameters Module.
 */
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
    ret_code_t err_code;

    if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
    {
        err_code = sd_ble_gap_disconnect(m_connection_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK(err_code);
    }
}

/**@brief Function for handling a Connection Parameters error.
 *
 * @param[in]   nrf_error   Error code containing information about what went wrong.
 */
static void conn_params_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}


/**@brief Function for initializing the Connection Parameters module.
 */
static uint32_t conn_params_init(void)
{
    ble_conn_params_init_t cp_init;

    memset(&cp_init, 0, sizeof(cp_init));

    cp_init.p_conn_params                  = &gap_conn_params;
    cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
    cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
    cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail             = false;
    cp_init.evt_handler                    = on_conn_params_evt;
    cp_init.error_handler                  = conn_params_error_handler;

    // Set the connection params in stack
    return sd_ble_gap_ppcp_set(cp_init.p_conn_params);
}

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @param[in] p_ble_evt  Bluetooth stack event.
 */
static bool on_generic_ble_evt(ble_evt_t * p_ble_evt)
{
    ret_code_t err_code = NRF_SUCCESS;

    switch (p_ble_evt->header.evt_id)
    {
        // GHS - Should never happen
        case BLE_EVT_USER_MEM_REQUEST:
        {
            err_code = sd_ble_user_mem_reply(m_connection_handle, NULL);
            APP_ERROR_CHECK(err_code);
        }
        break; // BLE_EVT_USER_MEM_REQUEST

        case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
        {
            ble_gatts_evt_rw_authorize_request_t  req;
            ble_gatts_rw_authorize_reply_params_t auth_reply;

            req = p_ble_evt->evt.gatts_evt.params.authorize_request;

            if (req.type != BLE_GATTS_AUTHORIZE_TYPE_INVALID)
            {
                if ((req.request.write.op == BLE_GATTS_OP_PREP_WRITE_REQ)     ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_NOW) ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL))
                {
                    if (req.type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_WRITE;
                    }
                    else
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
                    }
                    auth_reply.params.write.gatt_status = APP_FEATURE_NOT_SUPPORTED;
                    err_code = sd_ble_gatts_rw_authorize_reply(p_ble_evt->evt.gatts_evt.conn_handle,
                                                               &auth_reply);
                    APP_ERROR_CHECK(err_code);
                }
            }
        } 
        break; // BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST

        default:
            // No implementation needed.
            return false;
    }
    return true;
}

static bool handleCommonEvents(ble_evt_t* p_ble_evt, uint16_t mtu_size)
{
    ret_code_t err_code;
    bool used = false;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:
            mtu_size = p_ble_evt->evt.gatts_evt.params.exchange_mtu_request.client_rx_mtu;
            NRF_LOG_DEBUG("New MTU size requested %d", mtu_size);
            if (mtu_size > NRF_SDH_BLE_GATT_MAX_MTU_SIZE)   // Set in sdk_config.h
            {
                mtu_size = NRF_SDH_BLE_GATT_MAX_MTU_SIZE;
                NRF_LOG_DEBUG("Requested MTU size exceeds maximum; reset to maximum size of %d", mtu_size);
            }
            err_code = sd_ble_gatts_exchange_mtu_reply(p_ble_evt->evt.gatts_evt.conn_handle, mtu_size);

            if (err_code != NRF_SUCCESS)
            {
                mtu_size = NRF_SDH_BLE_GATT_MAX_MTU_SIZE;
                NRF_LOG_ERROR("MTU exchange request reply failed. Error: %s (0x%04X)", getErrorString(err_code), err_code);
            }
            used = true;
            break;

        case BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST:
        {
            NRF_LOG_DEBUG("Data length update request: max_rx_octets %d, max_rx_time %d, "
                "max_tx_octets %d, max_tx_time %d",
                p_ble_evt->evt.gap_evt.params.data_length_update_request.peer_params.max_rx_octets,
                p_ble_evt->evt.gap_evt.params.data_length_update_request.peer_params.max_rx_time_us,
                p_ble_evt->evt.gap_evt.params.data_length_update_request.peer_params.max_tx_octets,
                p_ble_evt->evt.gap_evt.params.data_length_update_request.peer_params.max_tx_time_us);
            uint32_t result = sd_ble_gap_data_length_update(
                p_ble_evt->evt.gap_evt.conn_handle,
                NULL, //&p_ble_evt->evt.gap_evt.params.data_length_update_request.peer_params,
                NULL);
            //NRF_LOG_DEBUG("Result %d", result);
            used = true;
        }
        break;

        case BLE_GAP_EVT_DATA_LENGTH_UPDATE:
        {
            NRF_LOG_DEBUG("Data length update: max_rx_octets %d, max_rx_time %d, "
                "max_tx_octets %d, max_tx_time %d",
                p_ble_evt->evt.gap_evt.params.data_length_update.effective_params.max_rx_octets,
                p_ble_evt->evt.gap_evt.params.data_length_update.effective_params.max_rx_time_us,
                p_ble_evt->evt.gap_evt.params.data_length_update.effective_params.max_tx_octets,
                p_ble_evt->evt.gap_evt.params.data_length_update.effective_params.max_tx_time_us);
        }
        used = true;
        break;

        case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
        {
            NRF_LOG_DEBUG("Physical update request: rx phys %d, tx phys %d",
                p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys.rx_phys,
                p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys.tx_phys);
            const ble_gap_phys_t* phys = &p_ble_evt->evt.gap_evt.params.phy_update_request.peer_preferred_phys;
            uint32_t result = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle,
                phys);
            //NRF_LOG_DEBUG("Result %d", result);
            used = true;
        }

        case BLE_GAP_EVT_PHY_UPDATE:
        {
            //NRF_LOG_DEBUG("Physical update update:rx phys %d, tx phys %d,",
            //    p_ble_evt->evt.gap_evt.params.phy_update.rx_phy,
            //    p_ble_evt->evt.gap_evt.params.phy_update.tx_phy);
            used = true;
        }
        break;
    }
    return used;
}

static ret_code_t notify_msmt(unsigned char *msmt_char, unsigned short handle, unsigned short hvx_length)
{
    ret_code_t             error_code;
    ble_gatts_hvx_params_t hvx_params;
    char buffer[32];
    memset(buffer, 0, 32);
    NRF_LOG_INFO("=====> Sending %u bytes of data %s at time %u:", 8, (uint32_t)byteToHex(msmt_char, buffer, " ", 8), getTicksMillis());
    NRF_LOG_FLUSH();

    if (cycle_abort // connection has been terminated by PHG
        || (m_connection_handle == BLE_CONN_HANDLE_INVALID))    // the connection has been killed
    {
        NRF_LOG_DEBUG("=====> Aborted or connection handle invalid");
        sd_mutex_release(&p_mutex);
        emptyQueue(queue);
        return NRF_ERROR_INVALID_STATE;
    }
    hvx_params.handle = handle;
    hvx_params.type = BLE_GATT_HVX_NOTIFICATION;
    hvx_params.offset = 0; // global_send_offset;
    hvx_params.p_len = &hvx_length;
    hvx_params.p_data = msmt_char;

    // Send notification
    error_code = sd_ble_gatts_hvx(m_connection_handle, &hvx_params);
    if (error_code == NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("=====> Entire package sent");
        sd_mutex_release(&p_mutex);
        return NRF_SUCCESS;
    }
    // If we have no buffers, we have to wait for the BLE_EVT_TX_COMPLETE event and resend
    else if( error_code == NRF_ERROR_RESOURCES) // Notifications only
    {
        NRF_LOG_DEBUG("=====> No TX buffers; wait for event and resend.");
        sd_mutex_release(&p_mutex);
        return NRF_ERROR_RESOURCES;
    }
    // Hopefully this does not happen.
    else
    {
        NRF_LOG_ERROR("=====> Failed doing the notification. Error: %s (0x%04X)", getErrorString(error_code), error_code);
        sd_mutex_release(&p_mutex);
        emptyQueue(queue);
        return error_code;
    }
}

static ret_code_t send_data()
{
    ret_code_t             error_code = NRF_SUCCESS;
    if(sd_mutex_acquire(&p_mutex) == NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN)
    {
        NRF_LOG_DEBUG("Mutex locked");
        return NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN;
    }
    if (send_flag)
    {
        send_flag = false;
        if (cccdSet[0] != 0)
        {
            error_code = notify_msmt(cycle_power_msmt_char, m_cyc_pwr_msmt_char_handle.value_handle, 8);
            if ((error_code == NRF_SUCCESS) && (cccdSet[1] != 0))
            {
                error_code = notify_msmt(heart_rate_msmt_char, m_heart_rate_msmt_char_handle.value_handle, 3);
            }
        }
        NRF_LOG_INFO("Msmt notified, response: %s (0x%04X)", getErrorString(error_code), error_code);
    }
    sd_mutex_release(&p_mutex);
    return error_code;
}

static bool encodeLiveData(void *data)
{
    if (cccdSet[0] && !send_flag)
    {
        s_PowerMsmt *powerMsmt = (s_PowerMsmt *)data;

        cycle_power_msmt_char[2] = (powerMsmt->power & 0xFF);
        cycle_power_msmt_char[3] = ((powerMsmt->power >> 8) & 0xFF);
        unsigned long crankTime = (powerMsmt->crankTime < time_of_prev_last_crank) ? (65536 + powerMsmt->crankTime) :
                                        powerMsmt->crankTime;
        //unsigned long deltaT = crankTime - time_of_prev_last_crank;
        unsigned long revs = 0;
       // unsigned long diff = deltaT * powerMsmt->rpm - revs * 614400;  // solve for diff = 0 to get crankT
        unsigned long crankT = time_of_prev_last_crank;
        if (powerMsmt->rpm >  10)  // 1 rpm is the slowest we can handle
        {
            revs = (crankTime - time_of_prev_last_crank) * powerMsmt->rpm / 614400;        // (600 * 1024)
            crankT = crankT + (revs * 614400 / powerMsmt->rpm);
        }
        else
        {
            crankT = powerMsmt->crankTime;
        }

        cumulative_crank_revs = revs + cumulative_crank_revs;
        cycle_power_msmt_char[0] = CP_MSMT_FLAGS_HAS_CRANK_REVS;
        cycle_power_msmt_char[4] = (cumulative_crank_revs & 0xFF);
        cycle_power_msmt_char[5] = ((cumulative_crank_revs >> 8) & 0xFF);
        cycle_power_msmt_char[6] = (crankT  & 0xFF);
        cycle_power_msmt_char[7] = ((crankT >> 8) & 0xFF);

        NRF_LOG_INFO("revs %u cum crank revs %u crankT %u  prevT %u", revs, cumulative_crank_revs, crankT, time_of_prev_last_crank);
        time_of_prev_last_crank = crankT;

        if (cccdSet[1] != 0)
        {
            powerMsmt->hr = powerMsmt->hr / 10;
            heart_rate_msmt_char[1] = (powerMsmt->hr & 0xFF);
            heart_rate_msmt_char[2] = ((powerMsmt->hr >> 8) & 0xFF);
        }
        return true;
    }
    else
    {
        if (!cycle_abort)
        {
            emptyQueue(queue);
        }
        else
        {
            NRF_LOG_DEBUG("Not ready for measurement at time %lu", getTicksMillis());
        }
    }
    return false;
}

    /*  Keiser cycle non connectable adv (Data is in manufacturer data)
     *  03 09 4d 33  // full name
     *  03 19 00 00  // device class
     *  02 01 04     // flags
     *  14 ff        // manuf
        Byte Index  Attribute           Size    Example Value
        0, 1        Company ID          2 Byte  02 01
        2           Version Major       1 Byte  06
        3           Version Minor       1 Byte  30
        4           Data Type           1 Byte  00
        5           Equipment ID        1 Byte  38
        6, 7        Cadence             2 Byte  38 03   rpm X 10
        8, 9        Heart Rate          2 Byte  46 05   bpm X 10
        10, 11      Power               2 Byte  73 00   watts  - error in documentation and simulator Its just watts
        12, 13      Caloric Burn        2 Byte  0D 00
        14          Duration Minutes    1 Byte  04
        15          Duration Seconds    1 Byte  27
        16, 17      Distance            2 Byte  01 00   miles or km X 10 high order bit 1 = km, 0 = miles
        18          Gear                1 Byte  0A

     */
static bool isKeiser(unsigned char *adv, int tot_len)
{
    int index = 0;
    while(index < tot_len)
    {
        int len = (adv[index] & 0xFF);
        index++;
        if(adv[index] != 0xFF)  // Look for manufacturer data
        {
            index = index + len;    // If not, skip to next ad type's length
        }
        else
        {
            if (len >= 0x14 && adv[index + 1] == 2 && adv[index + 2] == 1)  // Keiser has 0x02, 0x01 identfier and has at least a length of 20 bytes
            {
                first_keiser_adv_found = true;
                unsigned short mode = (unsigned short)adv[index + 5];
                if ((unsigned short)adv[index + 4] < 30)    // minor version less thn 30 had adv frequency about 2 seconds
                {
                    mask = 0;   // Set mask to zero, no average, every adv converted and sent
                }
                if (mode == 0 || (mode >= 128 && mode <= 227))  // Real time mode 0 or  128 - 227; all else is review mode
                {
                    heart_rate = heart_rate + (adv[index + 9] & 0xFF) + ((adv[index + 10] << 8) & 0xFF00); // heart rate
                    power = power + (adv[index + 11] & 0xFF) + ((adv[index + 12] << 8) & 0xFF00); // power
                    rpm =     rpm + (adv[index +  7] & 0xFF) + ((adv[index +  8] << 8) & 0xFF00); // RPM times 10
                }
                else
                {
                    power = 0;
                    rpm = 0;
                    heart_rate = 0;
                }
                return true;
            }
            return false;
        }
    }
    return false;
}
static void bring_up_adver(void);      // Needed in disconnect handler
static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
    uint32_t              err_code;
    ble_gap_conn_params_t connParams;

    if (p_ble_evt == NULL)
    {
        return;
    }

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
        {
            NRF_LOG_INFO("Connection event received at time %u with saveDataLength = %u.", getTicksMillis(), saveDataLength);
            m_connection_handle = p_ble_evt->evt.gap_evt.conn_handle;
            /*
            if (saveDataBuffer != NULL)
            {
                // Calling sd_ble_gatts_sys_attr_set with CCCD info
                err_code = sd_ble_gatts_sys_attr_set(p_ble_evt->evt.gap_evt.conn_handle, saveDataBuffer,
                    saveDataLength, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS | BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
                if (err_code != NRF_SUCCESS)
                {
                    NRF_LOG_ERROR("Failed updating persistent sys attr info. Error: %s (0x%04X)", getErrorString(err_code), err_code);
                    free(saveDataBuffer);
                    saveDataBuffer = NULL;
                    saveDataLength = 0;
                }
            }
            */
            // start scanning for Keiser nonconnectable advertisments
            NRF_LOG_INFO("Starting scanner");

            bsp_board_led_on(BLUE_LED);
            bsp_board_led_off(ADVERTISING_LED);

            err_code = sd_ble_gap_scan_start(&scan_params, &scan_buffer_t);
            if (err_code != NRF_SUCCESS)
            {
                NRF_LOG_ERROR("Error starting scanner. Error: %s (0x%04X)", getErrorString(err_code), err_code);
            }
            cumulative_crank_revs = 0;
            power = 0;
            rpm = 0;
            heart_rate = 0;
            adv_count = 1;  // assure it starts at one
            elapsedCountStart = 0;
            elapsedCountStart = getTicks();
        }
        break;

        // Here is where one receives advertisement data
        case BLE_GAP_EVT_ADV_REPORT:
        {
            if (p_ble_evt->evt.gap_evt.params.adv_report.type.connectable == 0 &&   // non-Connectable
                p_ble_evt->evt.gap_evt.params.adv_report.type.scan_response == 0)
            {
                unsigned char *adv = p_ble_evt->evt.gap_evt.params.adv_report.data.p_data;
                if (isKeiser(adv, p_ble_evt->evt.gap_evt.params.adv_report.data.len))
                {

                    if ((adv_count & mask) == 0)    // For versions less than 6.30, mask is 0; every measurement is sent
                    {
                        char printBuf[256];
                        memset(printBuf, 0, 256);
                        NRF_LOG_INFO("Received Keiser advertisement %s at time %u", 
                            (uint32_t)byteToHex(adv, printBuf, " ", p_ble_evt->evt.gap_evt.params.adv_report.data.len),
                             getTicksMillis());
                        NRF_LOG_FLUSH();
                        s_PowerMsmt powerMsmt;
                        powerMsmt.hr = (heart_rate + (adv_count >> 1)) / adv_count;
                        powerMsmt.power = (power + (adv_count >> 1)) / adv_count;
                        powerMsmt.rpm  = (rpm + (adv_count >> 1)) / adv_count;      // Don't divide by 10 here, done in the crank calulations
                        powerMsmt.crankTime = (unsigned short)getTicks();
                        bsp_board_led_on(RED_LED);
                        NRF_LOG_INFO("Msmt added to queue");
                        enqueue(queue, &powerMsmt, sizeof(s_PowerMsmt));
                        power = 0;
                        rpm = 0;
                        heart_rate = 0;
                        adv_count = 0; // Must start at 1, gets advanced to 1 next line
                    }
                    adv_count++;
                }
            }
            sd_ble_gap_scan_start(NULL, &scan_buffer_t);
        }
        break;

        // Service change indication acknolwedged
        case BLE_GATTS_EVT_SC_CONFIRM:
        {
            NRF_LOG_DEBUG("Service change acknowledged at time %u", getTicksMillis());
        }
        break;


        // Added just for info
        case BLE_GAP_EVT_CONN_PARAM_UPDATE:
        {
            connParams = p_ble_evt->evt.gap_evt.params.conn_param_update.conn_params;
            NRF_LOG_DEBUG("Connection parameters update at time %u: min: %d max: %d supervision timeout: %d slave latency: %d ",
                getTicksMillis(),
                connParams.min_conn_interval, connParams.max_conn_interval, 
                connParams.conn_sup_timeout, connParams.slave_latency);
            break;
                
            case BLE_GATTC_EVT_TIMEOUT:
            NRF_LOG_DEBUG("GATT Client Timeout.");
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
        }
        break;


        case BLE_GATTS_EVT_TIMEOUT:
        {
            // Disconnect on GATT Server timeout event.
            NRF_LOG_DEBUG("GATT Server Timeout.");
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
        }
        break;


        case BLE_GAP_EVT_DISCONNECTED:
        {
//            uint16_t currentSysDataLength;
//            unsigned char *currentSysDataBuffer;
            bsp_board_led_off(BLUE_LED);
            bsp_board_led_off(ADVERTISING_LED);
            bsp_board_led_on(DISCONNECTED_LED);
            NRF_LOG_INFO("Disconnected at time %u", getTicksMillis());
            cycle_abort = true;
            sd_ble_gap_scan_stop();

            /*
            // call twice; once to get the size of the data
            // create the buffer,
            // call a second time to load the data into the buffer
            sd_ble_gatts_sys_attr_get(p_ble_evt->evt.gap_evt.conn_handle, NULL, &currentSysDataLength,
                BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS | BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
            // Do we need to upDate flash?
            if (currentSysDataLength > 0) // THis should always have something!!
            {
                // Lets get the current system info
                currentSysDataBuffer = calloc(1, currentSysDataLength * sizeof(uint8_t));
                err_code = sd_ble_gatts_sys_attr_get(p_ble_evt->evt.gap_evt.conn_handle, currentSysDataBuffer, &currentSysDataLength,
                    BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS | BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
                if (err_code != NRF_SUCCESS)
                {
                    NRF_LOG_ERROR("Failed getting persistent sys attr info. Error: %s (0x%04X)", getErrorString(err_code), err_code);
                }
                bool isEqual = ((currentSysDataLength == saveDataLength)
                                && (saveDataLength > 0 && saveDataBuffer != NULL)
                                && (memcmp(saveDataBuffer, currentSysDataBuffer, saveDataLength) == 0));
                if (isEqual)
                {
                    NRF_LOG_INFO("Flash write not needed; data is equal to what is already in flash");
                    flash_write_needed = false;
                }
                else
                {
                    if (saveDataBuffer != NULL)
                    {
                        free (saveDataBuffer);
                    }
                    saveDataBuffer = calloc (1, currentSysDataLength *sizeof(uint8_t));
                    saveDataLength = currentSysDataLength;
                    memcpy(saveDataBuffer, currentSysDataBuffer, saveDataLength);
                }
                free(currentSysDataBuffer);
            }
            */

            m_connection_handle = BLE_CONN_HANDLE_INVALID;

            /*
            if (flash_write_needed)
            {
                // this will disable soft device causing app to shut down after flash is written
                saveKeysToFlash(&keys, &saveDataBuffer, &saveDataLength, cccdSet, &noOfCccds);
                // Now LEDs wont work - everything is dead due to the sd_softdevice_disable.
                if (saveDataBuffer != NULL)
                {
                    free(saveDataBuffer);
                }
                break;
            }
            */

            #if (USE_DK == 0)
                restartAdv = true;
            #endif
        }
        break;


        case BLE_GAP_EVT_TIMEOUT:
        {
            NRF_LOG_DEBUG("Advertisement timed out at time %u", getTicksMillis());
        }
        break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
        {
            NRF_LOG_DEBUG("Saved info being requested at time %u", getTicksMillis());
            err_code = sd_ble_gatts_sys_attr_set(p_ble_evt->evt.gap_evt.conn_handle, saveDataBuffer,
                saveDataLength, BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS | BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
            if (err_code != NRF_SUCCESS)
            {
                if (saveDataBuffer != NULL)
                {
                    free (saveDataBuffer);
                    saveDataBuffer = NULL;
                    saveDataLength = 0;
                }
                NRF_LOG_ERROR("Failed updating persistent sys attr info. Error: %s (0x%04X)", getErrorString(err_code), err_code);
            }
        }
        break;

        /*  Pairing - encryption sequence:
         *  If not paired:
         *      BLE_GAP_EVT_SEC_PARAMS_REQUEST
         *      BLE_GAP_EVT_CONN_SEC_UPDATE
         *      BLE_GAP_EVT_AUTH_STATUS
         *  If paired:
         *      BLE_GAP_EVT_SEC_INFO_REQUEST
         *      BLE_GAP_EVT_CONN_SEC_UPDATE
         */

        // Pairing request from the PHG
        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
        {
            NRF_LOG_DEBUG("Pairing request requested");
            if (keys.keys_own.p_enc_key->enc_info.ltk_len != 0)     // Pairing request from a different PHG or PHG lost its bonding info.
            {
                NRF_LOG_DEBUG("GHS is paired with some device but not the one that has connected!");
                if (saveDataBuffer != NULL)
                {
                    free (saveDataBuffer);
                    saveDataLength = 0;
                    saveDataBuffer = NULL;
                }
            }
            memcpy(&peer_sec_params, &p_ble_evt->evt.gap_evt.params.sec_params_request.peer_params, sizeof(ble_gap_sec_params_t));
            err_code = sd_ble_gap_sec_params_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_SEC_STATUS_SUCCESS, &own_sec_params, &keys);
            APP_ERROR_CHECK(err_code);
        }
        break;

        // Encryption has been established
        case BLE_GAP_EVT_CONN_SEC_UPDATE:
        {
            NRF_LOG_DEBUG("Encryption established");
            hasEncrypted = true;
            bsp_board_led_on(ADVERTISING_LED);
            bsp_board_led_on(RED_LED);
        }
        break;

        /* This event is signaled when pairing has completed. All the key information is
         * placed in the 'keys' struct passed into the sd_ble_gap_sec_params_reply() method.
         * The next event to be signaled will be encryption.
         * If already paired, this event will not be signaled
         */
        case BLE_GAP_EVT_AUTH_STATUS:
        {
            NRF_LOG_DEBUG("Pairing completed");
        }
        break;
        
        // For requesting pass keys - should not happen here.
        case BLE_GAP_EVT_AUTH_KEY_REQUEST:
        {
            uint32_t result = sd_ble_gap_auth_key_reply(p_ble_evt->evt.gap_evt.conn_handle, BLE_GAP_AUTH_KEY_TYPE_NONE, NULL);
        }
        break;

        /* This event is signaled when encryption is being requested. Should only happen on a bonded reconnect. 
         * Here the keys that are needed are those of the peripheral. The keys of the slave device are used */
        case BLE_GAP_EVT_SEC_INFO_REQUEST:
        {
            NRF_LOG_DEBUG("Encryption request received");
            flash_write_needed = false; // We already have our data.
            ble_gap_evt_sec_info_request_t sec_info = p_ble_evt->evt.gap_evt.params.sec_info_request;
            ble_gap_enc_info_t* enc_info = NULL;
            ble_gap_irk_t* irk_info = NULL;

            if (sec_info.enc_info) // Set if peer is asked for the encryption info
            {
                enc_info = &keys.keys_own.p_enc_key->enc_info; // Keys created by the peripheral are used
            }
            if (sec_info.id_info) // Set if peer is asking for the IRK
            {
                irk_info = &keys.keys_own.p_id_key->id_info;
            }
            uint32_t result = sd_ble_gap_sec_info_reply(p_ble_evt->evt.gap_evt.conn_handle, enc_info, irk_info, NULL);
        }
        break;


        case BLE_GATTS_EVT_HVC:
            break;

        // By the time this is signaled, the write response has already been sent - so one is free to do whatever.
        case BLE_GATTS_EVT_WRITE:
            // Enable IND on CP
            // Enable NOTIFY power measurement
            if (p_ble_evt->evt.gatts_evt.params.write.handle == m_cyc_pwr_msmt_char_handle.cccd_handle)
            {
                uint8_t write_data = p_ble_evt->evt.gatts_evt.params.write.data[0];
                cccdSet[0] = (write_data == BLE_GATT_HVX_NOTIFICATION);
                NRF_LOG_INFO("Enabling Cycle Power Measurement CCCD with %d at time %u", write_data, getTicksMillis());
            }
            // Enable NOTIFY heart rate measurement
            if (p_ble_evt->evt.gatts_evt.params.write.handle == m_heart_rate_msmt_char_handle.cccd_handle)
            {
                uint8_t write_data = p_ble_evt->evt.gatts_evt.params.write.data[0];
                cccdSet[1] = (write_data == BLE_GATT_HVX_NOTIFICATION);
                NRF_LOG_INFO("Enabling Heart Rate Measurement CCCD with %d at time %u", write_data, getTicksMillis());
            }
            break;

        // Turns out for notifications one sends in a loop and you do get one event per N packets. Could be one event per notification
        // of one event per N notifications. So I need to keep track of outstanding events. global_send.chunks_outstanding is incremented
        // every notification, and decremented by p_ble_evt->evt.common_evt.params.tx_complete.count every event. When 0, the notification
        // sequence is done.
        case BLE_GATTS_EVT_HVN_TX_COMPLETE:  // This is the best we get for notifications
            NRF_LOG_INFO("A Measurement notification sent");
            break;


        default:
        {
            if (on_generic_ble_evt(p_ble_evt))
            {
                break;
            }
            if (handleCommonEvents(p_ble_evt, NRF_SDH_BLE_GATT_MAX_MTU_SIZE)) break;
            NRF_LOG_DEBUG("Unhandled Event %d at time %u", p_ble_evt->header.evt_id, getTicksMillis());
        }
        break;

        }
}

static void write_flash(void * p_context)
{
    saveKeysToFlash(&keys, &saveDataBuffer, &saveDataLength, cccdSet, &noOfCccds);
}

#define RAM_START       0x20000000
static uint32_t ram_end_address_get(void)
{
    uint32_t ram_total_size;

#ifdef NRF51
    uint32_t block_size = NRF_FICR->SIZERAMBLOCKS;
    ram_total_size      = block_size * NRF_FICR->NUMRAMBLOCK;
#else
    ram_total_size      = NRF_FICR->INFO.RAM * 1024;
#endif

    return RAM_START + ram_total_size;
}

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    ret_code_t err_code;

    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(err_code);

    uint32_t const app_ram_start_link = ram_start;
    err_code = sd_ble_enable(&ram_start); // This can change the ram start I guess.
    if (ram_start > app_ram_start_link)
    {
        NRF_LOG_WARNING("Insufficient RAM allocated for the SoftDevice.");

        NRF_LOG_WARNING("Change the RAM start location from 0x%x to 0x%x.",
                        app_ram_start_link, ram_start);
        NRF_LOG_WARNING("Maximum RAM size for application is 0x%x.",
                        ram_end_address_get() - (ram_start));
    }
    else
    {
        NRF_LOG_INFO("RAM starts at 0x%x", app_ram_start_link);
        if (ram_start != app_ram_start_link)
        {
            NRF_LOG_INFO("RAM start location can be adjusted to 0x%x.", ram_start);

            NRF_LOG_INFO("RAM size for application can be adjusted to 0x%x.",
                          ram_end_address_get() - (ram_start));
        }
    }
}

// Not sure what this used for
static void gpiote_init_new(void)
{
   // APP_GPIOTE_INIT(APP_GPIOTE_MAX_USERS);
}

// This method is called only once after the final measurement is received from the
// sensor module. At this point we start advertising, connect and send the obtained 
// final measurements
static void bring_up_adver(void)
{
    if (m_connection_handle  == BLE_CONN_HANDLE_INVALID)  // Should always be invalid
    {
        emptyQueue(queue);
        bsp_board_led_off(RED_LED);
        bsp_board_led_off(BLUE_LED);
        bsp_board_led_on(ADVERTISING_LED);
        hasEncrypted = false;
        cycle_abort = false;
        first_keiser_adv_found = false;
        flash_write_needed = true;
        uint32_t err_code = app_timer_start(m_app_dummy_timer_id, CMD_SENSOR_TIME, NULL);
        APP_ERROR_CHECK(err_code);
        advertising_start();
    }
}

#if (USE_DK == 1)
/*  We only use on the DK in place of no real sensor. To use this code we have to
    bring back the Board Support files into the project. */
static void bsp_event_handler(bsp_event_t event)
{
    switch (event)
    {
        case BSP_EVENT_SLEEP:
        break;

        case BSP_EVENT_KEY_3:
        break;

        case BSP_EVENT_KEY_1:
        NRF_LOG_INFO("Sending disconnect");
        app_timer_start(m_ghs_disconnect_timer_id, GHS_COMMAND_DELAY, NULL);
        break;

        case BSP_EVENT_KEY_2:
        {
            NRF_LOG_INFO("Button press done to start advertising");
            bring_up_adver();
          //  app_timer_start(m_ghs_live_data_timer_id, GHS_PLETH_DELAY, NULL);
        }
        break;

        default:
        NRF_LOG_DEBUG("What Button press was this???");
        break;
    }
}

static void buttons_leds_init(void)
{
    bsp_event_t startup_event;

    ret_code_t err_code = bsp_init(BSP_INIT_LEDS | BSP_INIT_BUTTONS, bsp_event_handler);

    APP_ERROR_CHECK(err_code);

    err_code = bsp_btn_ble_init(NULL, &startup_event);
    APP_ERROR_CHECK(err_code);
}
#else
     /*
    // Added this USB init stuff from the connectivity app. Do I need it? Can the trigger library do it?
    #if defined(APP_USBD_ENABLED) && APP_USBD_ENABLED

        static volatile bool m_usb_started;


        static void usbd_user_evt_handler(app_usbd_event_type_t event)
        {
            switch (event)
            {
                case APP_USBD_EVT_DRV_SUSPEND:
                    break;
                case APP_USBD_EVT_DRV_RESUME:
                    break;
                case APP_USBD_EVT_STARTED:
                    m_usb_started = true;
                    break;
                case APP_USBD_EVT_STOPPED:
                    app_usbd_disable();
                    break;
                case APP_USBD_EVT_POWER_DETECTED:
                    NRF_LOG_INFO("USB power detected");

                    if (!nrf_drv_usbd_is_enabled())
                    {
                        app_usbd_enable();
                    }
                    break;
                case APP_USBD_EVT_POWER_REMOVED:
                    NRF_LOG_INFO("USB power removed");
                    app_usbd_stop();
                    break;
                case APP_USBD_EVT_POWER_READY:
                    NRF_LOG_INFO("USB ready");
                    app_usbd_start();
                    break;
                default:
                    break;
            }
        }

        static void usbd_init(void)
        {
            app_usbd_serial_num_generate();
            static const app_usbd_config_t usbd_config = {
                .ev_state_proc = usbd_user_evt_handler
            };
            APP_ERROR_CHECK(app_usbd_init(&usbd_config));
        }

        static void usbd_enable(void)
        {
            #ifdef BOARD_PCA10059
                APP_ERROR_CHECK(nrf_dfu_trigger_usb_init());
            #endif
            APP_ERROR_CHECK(app_usbd_power_events_enable()); 

            // Process USB events until USB is started. This is related to the fact that
            // current version of softdevice does not handle USB POWER events.
            while (m_usb_started == false)
            {
                while (app_usbd_event_queue_process())
                {
                    // Nothing to do
                }
            }

        }
    #endif //APP_USBD_ENABLED
    */
/**@brief Function for the LEDs initialization.
 *
 * @details Initializes all LEDs used by the application.
 */
static void buttons_leds_init(void)
{
    // Do I need to make this call? Can the bsp_board_init() do the task?
    //nrf_dfu_trigger_usb_init();
    bsp_board_init(BSP_INIT_LEDS);
}
#endif

static void initializeBluetooth()
{
    s_DisCharValues disCharValues;
    // Initialize the DIS
    memset(&disCharValues, 0, sizeof(s_DisCharValues));
    disCharValues.modelNumber = MODEL_NUMBER;
    disCharValues.manufacturerName = MANUFACTURER_NAME;
    memcpy(disCharValues.systemId, systemId, 8);
    disCharValues.firmwareRevision = "0.0.1";
    
    ble_stack_init();

    // Set device address
    ble_gap_addr_t addrStruct;
    addrStruct.addr_type = BLE_GAP_ADDR_TYPE_PUBLIC;
    memcpy(addrStruct.addr, bluetoothAddress, 6);
    uint32_t err_code = sd_ble_gap_addr_set(&addrStruct);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not set the Bluetooth Address");
        APP_ERROR_CHECK(err_code);
    }

    // Sets max and min connection intervals, slave latency, and the GAP characteristic entries
    // for the friendly name and appearance.
    gap_params_init();
    // ------------------------ Creates the advertisement and scan response data arrays
    err_code = advertisement_ghs_set(&m_adv_handle);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not configure the advertisements");
        APP_ERROR_CHECK(err_code);
    }
    // ===================================== Create the DIS
    err_code = createDeviceInformationService(&m_dev_info_service_handle, &disCharValues);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_ERROR("Could not create the device information service. Reason: %s (0x%04X)", getErrorString(err_code), err_code);
    }
    // ===================================== Create the Cycle Power service
    err_code = createPrimaryService(&m_cyc_pwr_service_handle, BTLE_CYCLE_POWER_SERVICE);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not create Cycle Power service");
        APP_ERROR_CHECK(err_code);
    }
    // ===================================== Create the Cycle Power Measurement characteristic
    err_code = createStandardCharacteristic(m_cyc_pwr_service_handle,
        &m_cyc_pwr_msmt_char_handle,
        BTLE_CYCLE_POWER_MSMT_CHAR,
        true,   // Has CCCD
        false,   // This will cause indicate instead of notify if CCCD is set to true
        0,
        NULL,
        false,
        (ble_gap_conn_sec_mode_t) {1, 1},   // [CCCD write is open]
        (ble_gap_conn_sec_mode_t) {0, 0},   // Reading characteristic value is forbidden
        (ble_gap_conn_sec_mode_t) {0, 0},  // Writing characteristic value is forbidden
        false);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not create Cycle Power Measurement characteristic");
        APP_ERROR_CHECK(err_code);
    }
    // ------------------------------------- Add CCCD descriptor to power measurement char
    ble_gatts_attr_t pw_desc;
    memset(&pw_desc, 0, sizeof(ble_gatts_attr_t));
    ble_gatts_attr_md_t  pw_md;
    //ble_gatts_attr_md_t* pw_md_ptr = NULL;
    ble_uuid_t pw_uuid;
    pw_uuid.uuid = BTLE_SCCD_DESCRIPTOR;
    pw_uuid.type = BLE_UUID_TYPE_BLE;
    pw_desc.p_uuid = &pw_uuid;
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&pw_md.read_perm);
    pw_md.write_perm = (ble_gap_conn_sec_mode_t) {1, 1};
    pw_md.vloc = BLE_GATTS_VLOC_STACK;
    //pw_md_ptr = &pw_md;
    pw_desc.p_attr_md = &pw_md;
    err_code = sd_ble_gatts_descriptor_add(m_cyc_pwr_msmt_char_handle.value_handle, &pw_desc, 
            &m_cyc_pwr_msmt_char_handle.sccd_handle);
    // ===================================== Create the Cycle Power Feature characteristic
    err_code = createStandardCharacteristic(m_cyc_pwr_service_handle,
        &m_cyc_pwr_feature_char_handle,
        BTLE_CYCLE_POWER_FEATURE_CHAR,
        false,   // Has no CCCD
        false,  // This will cause indicate instead of notify if CCCD is set to true
        4,
        feature,
        false,
        (ble_gap_conn_sec_mode_t){0, 0},
        // CCCD write is forbidden
        (ble_gap_conn_sec_mode_t){1, 1},
        // Reading characteristic value is open
        (ble_gap_conn_sec_mode_t){0, 0},  // Writing characteristic is forbidden
        true);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not create Cycle Power feature characteristic");
        APP_ERROR_CHECK(err_code);
    }
    err_code = createStandardCharacteristic(m_cyc_pwr_service_handle,
        &m_cyc_pwr_sensor_loc_char_handle,
        BTLE_SENSOR_LOCATION_CHAR,
        false,  // Has no CCCD
        false,  // This will cause indicate instead of notify if CCCD is set to true
        1,                // Value length
        sensorLocation,   // Value
        false,
        (ble_gap_conn_sec_mode_t){0, 0},
        // CCCD write is forbidden
        (ble_gap_conn_sec_mode_t){1, 1},
        // Reading characteristic value is open
        (ble_gap_conn_sec_mode_t){0, 0},  // Writing characteristic is forbidden
        true);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not create Sensor location characteristic");
        APP_ERROR_CHECK(err_code);
    }
    
    // ===================================== Create the HEART RATE service
    err_code = createPrimaryService(&m_heart_rate_service_handle, BTLE_HEART_RATE_SERVICE);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not create Heart Rate service");
        APP_ERROR_CHECK(err_code);
    }
    // ===================================== Create the Heart Rate Measurement characteristic
    err_code = createStandardCharacteristic(m_heart_rate_service_handle,
        &m_heart_rate_msmt_char_handle,
        BTLE_HEART_RATE_MSMT_CHAR,
        true,   // Has CCCD
        false,  // This will cause indicate instead of notify if CCCD is set to true
        0,
        NULL,
        false,
        (ble_gap_conn_sec_mode_t) {1, 1},  // [CCCD write is open]
        (ble_gap_conn_sec_mode_t) {0, 0},  // Reading characteristic value is forbidden
        (ble_gap_conn_sec_mode_t) {0, 0},  // Writing characteristic value is forbidden
        false);
    if (err_code != NRF_SUCCESS)
    {
        NRF_LOG_DEBUG("Could not create Heart Rate Measurement characteristic");
        APP_ERROR_CHECK(err_code);
    }
    // ------------------------------------- Add CCCD descriptor to heart rate measurement char
    ble_gatts_attr_t hr_desc;
    memset(&hr_desc, 0, sizeof(ble_gatts_attr_t));
    ble_gatts_attr_md_t  hr_md;
    //ble_gatts_attr_md_t* hr_md_ptr = NULL;
    ble_uuid_t hr_uuid;
    hr_uuid.uuid = BTLE_SCCD_DESCRIPTOR;
    hr_uuid.type = BLE_UUID_TYPE_BLE;
    hr_desc.p_uuid = &hr_uuid;
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&hr_md.read_perm);
    hr_md.write_perm = (ble_gap_conn_sec_mode_t) {1, 1};
    hr_md.vloc = BLE_GATTS_VLOC_STACK;
    //hr_md_ptr = &hr_md;
    hr_desc.p_attr_md = &hr_md;
    err_code = sd_ble_gatts_descriptor_add(m_heart_rate_msmt_char_handle.value_handle, &hr_desc, 
            &m_heart_rate_msmt_char_handle.sccd_handle);


    // Start execution.
    elapsedTimeStart = 0;
    elapsedTimeStart = getTicksMillis();
    NRF_LOG_INFO("Cycle Power Start at time %u!\r", getTicksMillis());
    #if (USE_DK == 0)
        bring_up_adver();
    #endif
}



/**@brief Function for the Power manager.
 */
static void power_manage(void)
{
    // I do not know where this is established. I don't know what
    // the 'app' event is that triggers this or where it setup.
    ret_code_t err_code = sd_app_evt_wait();
    
    if (err_code == NRF_SUCCESS)
    {
        //uint32_t evt_id;
        // Pull event from SOC.
        //err_code = sd_evt_get(&evt_id);
        //NRF_LOG_DEBUG("Soc event # %u", evt_id);
        return;
    }
    NRF_LOG_DEBUG("sd app evt wait error %u", err_code);
    APP_ERROR_CHECK(err_code);
}

void main_wait(void)
{
    if (NRF_LOG_PROCESS() == true)
    {
        return;
    }
    power_manage();
}

/**@brief Function for the main 'embedded' loop. The Nordic SoftDevice is event driven.
* Any method prefixed by sd_* is a SoftDevice method. SoftDevice methods consist of just
* your basic GATT and GAP level calls. The Nordic SDK provides methods to simplify certain
* processes but they all call these sd_* methods. Here we do not use the SDK but only sd_*.

* The key method in the wait loop is sd_app_evt_wait(). This method waits until a Bluetooth
* or system event is triggered. When that happens, the function returns. The application must
* then use the method sd_ble_evt_get() to get the Bluetooth Event that just occurred. It could
* be a read, write, connection, disconnection, etc. The application then must either handle
* the event or ignore it. In this application the event handler is ble_evt_dispatch(). Note
* that after the event is handled one has to call sd_ble_evt_get() again until the method
* returns that there are no more events. One then calls sd_app_evt_wait() and the process 
* repeats. But we extend that functionality.

* Note that the loop also contains a call to the application method send_data(). It is poorly
* named as it does both indications and notifications. What this method does is check if a flag
* has been set that data needs to be indicated/notified. If it has, the flag will be cleared and
* a hunk of data will be indicated/notified. If it is an indication, the app must wait until
* the indication is confirmed so the data is sent and the position in the buffer is updated by
* the amount of data sent. Then the method returns, and we wait for the confirmation event
* BLE_GATTS_EVT_HVC to occur. If all the segments have been sent, the next procedure occurs. If
* not, the indicate flag is set to true and the next segment is sent.

* If the data is notified, hunks can be sent one after the other without waiting for an event
* until the notification buffers are depleted. When that happens, the application must wait for
* the BLE_EVT_TX_COMPLETE. At that point we can continue. This loopy way of handling sending
* data segments would be equivalent to waiting on a semaphore in a real time OS.

* Recall that the GHS works as follows:
*   1. receive a command on the Control Point
*   2. if the command is to send data (current time, system info, config info, get stored records,
*      send live data) the data is notified on the response characteristic until all segments are
*      sent. 
*      Now if the data being sent is measurements, there may be several records, especially when
*      transferring stored data. A record consists of all the measurements with a common time
*      stamp. When a record is completed, the PHD server sends an indication on the CP saying a record
*      has completed. The PHG client shall not send another command yet. The PHD must send a second
*      indication on the CP sayong that all records have been sent. Then the PHG can send another
*      command.
*      The system info, current time info, and config info contain only one long entry. There is
*      no record in that case.
*   3. if the PHD does not support the command or the command is to get the number of stored
*      records or delete stored data, there is only an indication on the CP with the result or DONE.

* Pairing/bonding. This implementation requires that the client ask the PHDe if the PHD wants to
* pair. If the PHD responds 'yes' on the CP, the cleint begins pairing. To do this none of the
* characteristics or descriptors are secured. However, if the application wants to pair and the
* client does not, all other commands will fail. Probably should put a special error message like
* PHD requires pairing and the PHG did not pair. The device could also disconnect. This is up in
* the air. We do this to keep the agony of pairing as user-friendly as possible as well as synchronous.
* The client will never have to worry about getting a security request or insufficient autherntication
* error while doing something else. GREATLY simplifies cleint code not having to check for this crap.
* In addition, the user does not have to know if the PHD is pairable or unpairable; the user only has
* to discover and connect. Pairing has proved to be a user nightmare in the field.

* Clearly the command values and response codes are all up in the air. To just make this work
* a set of minimal values were chosen.
*/

static void main_loop(void)
{
    uint8_t enabled;
    uint16_t len;
    ret_code_t result;
    for (;;)
    {
        send_data();            // when flag is set, a set of data is indicated or notified depending upon setup. 
                                // Flag is reset in method
        main_wait();            // Contains the sd_app_evt_wait()
        if (!isEmpty(queue))
        {
            void *data = front(queue);
            if (encodeLiveData(data))
            {
                NRF_LOG_INFO("Measurement taken from queue");
                send_flag = true;
                dequeue(queue);
                bsp_board_led_off(RED_LED);
            }
        }
        if (restartAdv)
        {
            bsp_board_led_off(DISCONNECTED_LED);
            bsp_board_led_on(RED_LED);
            bsp_board_led_on(BLUE_LED);
            restartAdv = false;
            bring_up_adver();
        }
        while(true)
        {
            // uint32_t evt_id;

            // Don't continue on if disabled, for example when writing flash at the end
            sd_softdevice_is_enabled(&enabled);
            if (enabled != 1)
            {
                NRF_LOG_DEBUG("Restarting application");
                initializeBluetooth();
             //   while(NRF_LOG_PROCESS());
             //   NRF_POWER->RESETREAS = 0;
             //   NVIC_SystemReset();
             //   return;
            }
            result = sd_ble_evt_get(NULL, &len);    // Get size of event
            if (result == NRF_ERROR_NOT_FOUND)      // If there aren't any, go back to wait
            {
                break;
            }
            evt_buf = (uint8_t *)calloc(1, len);                // Make space for event. evt_buf is 4-byte aligned
            result = sd_ble_evt_get((uint8_t *)evt_buf, &len);  // get the event
            if (result == NRF_SUCCESS)
            {
                
                ble_evt_t *evt = (ble_evt_t *)evt_buf;
                ble_evt_dispatch(evt);                          // dispatch event to handler. This handles only BTLE related events
            }
            else                                                // Hopefully no error but just in case log it.
            {                                                   // Should I do an NVIC_SystemReset() here?
                NRF_LOG_DEBUG("PENDING BLE Event return error: %u", result);
                free(evt_buf);  // Clean up
                break;          // back to wait
            }
            free(evt_buf);      // clean up and get the next event
        }
    }
}

/**@brief Function for application main entry.
 */
int main(void)
{
    ret_code_t err_code;
    unsigned char cccds[2];

    // Initializing info for the scanner to receive Keiser advertisements
    memset(&scan_params, 0, sizeof(ble_gap_scan_params_t));
    scan_params.scan_phys = BLE_GAP_PHY_AUTO;
    scan_params.interval = NRF_BLE_SCAN_SCAN_INTERVAL;
    scan_params.window = NRF_BLE_SCAN_SCAN_WINDOW;
    scan_params.timeout = BLE_GAP_SCAN_TIMEOUT_UNLIMITED;

    memset(&scan_buffer, 0, sizeof(scan_buffer));
    scan_buffer_t.len = 31;
    scan_buffer_t.p_data = scan_buffer;


    m_adv_handle = 0;
    queue = initializeQueue(10);

    cycle_abort = false;

    // Initialize.
    err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);
    NRF_LOG_DEFAULT_BACKENDS_INIT();
    NRF_LOG_DEBUG("Main start Cycle Power Meter");


    sec_params_init();

    memset(&m_cyc_pwr_msmt_char_handle, 0, sizeof(m_cyc_pwr_msmt_char_handle));
    memset(&m_cyc_pwr_feature_char_handle, 0, sizeof(m_cyc_pwr_feature_char_handle));
    memset(&m_heart_rate_msmt_char_handle, 0, sizeof(m_heart_rate_msmt_char_handle));


    // Allocate memory for the security keys
//    allocateMemoryForSecurityKeys(&keys);
    memset(cccds, 0, noOfCccds);
//    loadKeysFromFlash(&keys, &saveDataBuffer, &saveDataLength, cccds, &noOfCccds);
//    memcpy(cccdSet, cccds, noOfCccds);


    timers_init();
    buttons_leds_init();
    bsp_board_led_on(RED_LED );
    bsp_board_led_on(BLUE_LED);
    
    sd_mutex_new(&p_mutex);

    gpiote_init_new();
    initializeBluetooth();
    main_loop();
}


/**
 * @}
 */
